import glob
import os

import keras_vgg_buddy
import numpy as np
from keras import backend as K
from scipy.misc import imsave

from .base import generate_img_batches


def input_generator(args):
    '''Generates batches of random samples paired with images.
    '''
    g_training_imgs = generate_img_batches(
        args.training_data_path, args.batch_size,
        resize_shape=(args.max_height, args.max_width))
    while True:
        data = {
            'content': np.array(next(g_training_imgs))
        }
        yield data


def evaluation_input_generator(args):
    '''Generates batches of random samples paired with images.
    '''
    g_training_imgs = generate_img_batches(
        args.eval_data_path, args.batch_size,
        resize_shape=(args.max_height, args.max_width))
    while True:
        data = {
            'content': np.array(next(g_training_imgs))
        }
        yield data


def transform_glob(model, args):
    '''Apply the model to a glob of images.'''
    f_generate = K.function([model.inputs['content'].input],
        [model.nodes['texnet'].get_output(False)])
    filenames = glob.glob(args.convert_glob)
    output_path = args.output_prefix
    try:
        os.makedirs(output_path)
    except OSError:
        pass  # exists
    for filename in filenames:
        print('converting {}'.format(filename))
        img = keras_vgg_buddy.load_and_preprocess_image(filename, width=args.max_width)
        result = f_generate([np.expand_dims(img, 0)])[0]
        img = keras_vgg_buddy.deprocess_image(result[0], contrast_percent=0)
        imsave(os.path.join(output_path, os.path.basename(filename)), img)


def output_size_from_glob(target_glob, width=256):
    filenames = glob.glob(target_glob)
    assert filenames, 'No files matched glob: {}'.format(target_glob)
    img0 = keras_vgg_buddy.load_and_preprocess_image(filenames[0], width=width)
    return img0.shape[-2:]



from keras import backend as K
from keras.regularizers import Regularizer


# the gram matrix of an image tensor (feature-wise outer product)
def gram_matrix(x):
    assert K.ndim(x) == 4
    xs = K.shape(x)
    features = K.reshape(x, (xs[0], xs[1], xs[2] * xs[3]))
    gram = K.batch_dot(features, K.permute_dimensions(features, (0, 2, 1)))
    return gram


class FeatureStyleRegularizer(Regularizer):
    '''Gatys et al 2015 http://arxiv.org/pdf/1508.06576.pdf'''
    def __init__(self, target=None, weight=1.0, num_inputs=1, **kwargs):
        self.target = target
        self.weight = weight
        self.num_inputs = num_inputs
        super(FeatureStyleRegularizer, self).__init__(**kwargs)

    def __call__(self, loss):
        output = self.layer.get_output(True)
        batch_size = K.shape(output)[0] // self.num_inputs
        generated = output[:batch_size, :, :, :]
        loss += self.weight * K.mean(
            K.sum(K.square(gram_matrix(self.target) - gram_matrix(generated)), axis=(1,2))
        ) / (4.0 * K.square(K.prod(K.shape(generated)[1:])))
        return loss


class FeatureContentRegularizer(Regularizer):
    '''Penalizes euclidean distance of content features.'''
    def __init__(self, weight=0.001, **kwargs):
        self.weight = weight
        super(FeatureContentRegularizer, self).__init__(**kwargs)

    def __call__(self, loss):
        output = self.layer.get_output(True)
        batch_size = K.shape(output)[0] // 2
        generated = output[:batch_size, :, :, :]
        content = output[batch_size:, :, :, :]
        loss += self.weight * K.mean(
            K.sum(K.square(content - generated), axis=(1,2,3))
        )
        return loss


class TVRegularizer(Regularizer):
    '''Enforces smoothness in image output.'''
    def __init__(self, weight=1.0, **kwargs):
        self.weight = weight
        super(TVRegularizer, self).__init__(**kwargs)

    def __call__(self, loss):
        x = self.layer.get_output(True)
        assert K.ndim(x) == 4
        a = K.square(x[:, :, 1:, :-1] - x[:, :, :-1, :-1])
        b = K.square(x[:, :, :-1, 1:] - x[:, :, :-1, :-1])
        loss += self.weight * K.mean(K.sum(K.pow(a + b, 1.25), axis=(1,2,3)))
        return loss


class MRFRegularizer(Regularizer):
    '''MRF loss http://arxiv.org/pdf/1601.04589v1.pdf'''
    def __init__(self, features, weight=1.0, patch_size=3, **kwargs):
        self.features = features
        self.weight = weight
        self.patch_size = patch_size
        super(MRFRegularizer, self).__init__(**kwargs)

    def __call__(self, loss):
        from . import patches

        output = self.layer.get_output(True)
        assert K.ndim(output) == 4
        batch_size = K.shape(output)[0] // 2
        patch_size = self.patch_size
        patch_stride = 1
        generated = output[batch_size:, :, :, :]
        # extract patches from feature maps
        generated_patches, generated_patches_norm = \
            patches.make_patches(generated, patch_size, patch_stride)
        target_patches, target_patches_norm = \
            patches.make_patches(self.features, patch_size, patch_stride)
        # find best patches and calculate loss
        patch_ids = patches.find_patch_matches(
            generated_patches, generated_patches_norm,
            target_patches / target_patches_norm)
        best_target_patches = K.reshape(
            target_patches[patch_ids], K.shape(generated_patches))
        loss += self.weight * K.sum(K.square(best_target_patches - generated_patches)) / patch_size ** 2
        return loss



'''Texture Network for style transfer.'''
import time

import keras_vgg_buddy
import numpy as np
from keras import backend as K
from keras.layers.core import Layer
from keras.layers.convolutional import AveragePooling2D
from keras.models import Graph
from keras.optimizers import Adam

from .base import create_res_texture_net, dumb_objective
from .regularizers import (
    FeatureContentRegularizer, FeatureStyleRegularizer, MRFRegularizer, TVRegularizer)


def make_model(args, style_img=None, activation='relu', pool_class=AveragePooling2D):
    model = Graph()
    model.add_input('content', batch_input_shape=(args.batch_size, 3, args.max_height, args.max_width))
    texnet = create_res_texture_net(args.max_height, args.max_width)
    # add the texture net to the model
    model.add_node(texnet, 'texnet', 'content')
    model.add_output('texture_rgb', 'texnet')
    # hook up the training network stuff
    if args.train:
        model.add_node(Layer(), 'vgg_concat', inputs=['texnet', 'content'], concat_axis=0)
        # add VGG and the constraints
        keras_vgg_buddy.add_vgg_to_graph(model, 'vgg_concat', pool_mode=args.pool_mode,
            trainable=False, weights_path=args.vgg_weights)
        # add the regularizers for the various feature layers
        vgg = keras_vgg_buddy.VGG16(args.max_height, args.max_width, pool_mode=args.pool_mode, weights_path=args.vgg_weights)
        print('computing static style features')
        style_features = vgg.get_features(np.expand_dims(style_img, 0), set(args.style_layers + args.content_layers + args.mrf_layers))
        regularizers = []
        if args.style_weight != 0.0:
            for layer_name in args.style_layers:
                layer = model.nodes[layer_name]
                style_regularizer = FeatureStyleRegularizer(
                    target=style_features[layer_name], weight=args.style_weight,
                    num_inputs=2)
                style_regularizer.set_layer(layer)
                regularizers.append(style_regularizer)
        if args.content_weight != 0.0:
            for layer_name in args.content_layers:
                layer = model.nodes[layer_name]
                content_regularizer = FeatureContentRegularizer(weight=args.content_weight)
                content_regularizer.set_layer(layer)
                regularizers.append(content_regularizer)
        if args.mrf_weight != 0.0:
            for layer_name in args.mrf_layers:
                layer = model.nodes[layer_name]
                mrf_regularizer = MRFRegularizer(K.variable(style_features[layer_name]), weight=args.mrf_weight)
                mrf_regularizer.set_layer(layer)
                regularizers.append(mrf_regularizer)
        if args.tv_weight != 0.0:
            tv_regularizer = TVRegularizer(weight=args.tv_weight)
            tv_regularizer.set_layer(model.nodes['texnet'])
            regularizers.append(tv_regularizer)
        setattr(model.nodes['vgg_concat'], 'regularizers', regularizers)  # Gotta put em somewhere?

    print('compiling')
    start_compile = time.time()
    adam = Adam(lr=args.learn_rate, beta_1=0.7)
    model.compile(optimizer=adam, loss=dict(texture_rgb=dumb_objective))
    print('Compiled model in {:.2f}'.format(time.time() - start_compile))
    return model



import glob
import os
import random
import time

import numpy as np
import keras_vgg_buddy
from keras import backend as K
from keras.callbacks import LearningRateScheduler


def generate_training_batches(args, input_generator):
    '''Prepare a batch of input data from input_generator to be used as training data.
    Essentially it's just satisfying the `dumb_objective` for `texture_rgb`
    '''
    dumb_output = dict(texture_rgb=np.zeros((args.batch_size, 3, args.max_width, args.max_width)))
    while True:
        data = next(input_generator)
        data.update(dumb_output)
        yield data


class TextureNetLearningSchedule(object):
    def __init__(self, batch_size, num_iterations_per_epoch, initial_lr=0.1, min_lr=0.001, cliff=10, falloff=0.9):
        self.batch_size = batch_size
        self.num_iterations_per_epoch = num_iterations_per_epoch
        self.initial_lr = initial_lr
        self.min_lr = min_lr
        self.cliff = cliff
        self.falloff = falloff

    def __call__(self, epoch):
        cliff_epoch = self.cliff # // self.num_iterations_per_epoch
        if epoch < cliff_epoch:  # run hot for a while
            new_lr = self.initial_lr
        else:
            new_lr = self.initial_lr * self.falloff ** ((epoch - cliff_epoch) + 1)
        new_lr = max(new_lr, self.min_lr)
        print('New learning rate: {}'.format(new_lr))
        return new_lr


def train(model, args, input_generator, callbacks=[], num_samples=10):
    lr_schedule = TextureNetLearningSchedule(
        args.batch_size, args.num_iterations_per_epoch,
        initial_lr=args.learn_rate, min_lr=args.min_learn_rate,
        cliff=20)
    callbacks.append(LearningRateScheduler(lr_schedule))
    try:
        model.fit_generator(
            generate_training_batches(args, input_generator),
            samples_per_epoch=args.num_iterations_per_epoch, nb_epoch=args.num_epochs,
            callbacks=callbacks
        )
    except KeyboardInterrupt:
        print('Stopping training...')



import os
import time

import numpy as np
import six
import keras.initializations
import keras_vgg_buddy
from keras import backend as K
from keras.layers.advanced_activations import LeakyReLU
from keras.layers.core import Activation, Layer
from keras.layers.convolutional import AveragePooling2D, Convolution2D, MaxPooling2D, UpSampling2D, ZeroPadding2D
from keras.layers.normalization import BatchNormalization
from keras.models import Graph


def add_conv_block(net, name, input_name, filters, filter_size, activation='relu', subsample=(1, 1)):
    net.add_node(Convolution2D(filters, filter_size, filter_size, subsample=subsample, border_mode='same'),
        name + '_conv', input_name)
    net.add_node(BatchNormalization(mode=0, axis=1), name + '_bn',  name + '_conv')
    net.add_node(Activation(activation), name,  name + '_bn')


def create_res_texture_net(input_rows, input_cols, num_res_filters=128, res_out_activation='linear', num_res_blocks=5):
    net = Graph()
    net.add_input('x', input_shape=(3, input_rows, input_cols))
    add_conv_block(net, 'in0', 'x', num_res_filters // 4, 9)
    add_conv_block(net, 'in1', 'in0', num_res_filters // 2, 3, subsample=(2, 2))
    add_conv_block(net, 'in2', 'in1', num_res_filters, 3, subsample=(2, 2))
    last_block_name = 'in2'
    for res_i in range(num_res_blocks):
        block_name = 'res_{}'.format(res_i)
        add_conv_block(net, block_name + '_in0', last_block_name, num_res_filters, 3)
        add_conv_block(net, block_name + '_in1', block_name + '_in0', num_res_filters, 3, activation='linear')
        net.add_node(Activation(res_out_activation), block_name, merge_mode='sum', inputs=[block_name + '_in1', last_block_name])
        last_block_name = block_name
    # theano doesn't seem to support fractionally-strided convolutions at the moment
    net.add_node(UpSampling2D(), 'out_up0', last_block_name)
    add_conv_block(net, 'out_0', 'out_up0', num_res_filters // 2, 3)
    net.add_node(UpSampling2D(), 'out_up1', 'out_0')
    add_conv_block(net, 'out_1', 'out_up1', num_res_filters // 4, 3)
    add_conv_block(net, 'out_2', 'out_1', 3, 9, activation='linear')
    net.add_node(Activation('linear'), 'texture_rgb', 'out_2', create_output=True)
    return net


def dumb_objective(x, y):
    '''Returns 0 in a way that makes everyone happy.

    Keras requires outputs and objectives but we're training purely upon the
    loss expressed by the regularizers.
    '''
    return 0.0 * y + 0.0 * x



import importlib


MODEL_CONFIG = {
    'transfer': {
        'input_generator': 'keras_rtst.generators.style_xfer',
        'evaluation_input_generator': 'keras_rtst.generators.style_xfer',
        'make_model': 'keras_rtst.models.style_xfer'
    }
}


def get_model_by_name(name):
    conf = MODEL_CONFIG[name]
    results = []
    output_order = ('make_model', 'input_generator', 'evaluation_input_generator')
    for prop in output_order:
        value = getattr(importlib.import_module(conf[prop]), prop)
        results.append(value)
    return results



import os
import random

import keras_vgg_buddy
import numpy as np
from keras import backend as K
from scipy.misc import imsave


def output_samples(model, args, input_generator):
    '''Generate args.num_sample outputs drawn from the `input_generator`.'''
    num_samples = args.num_samples
    num_batches = max(num_samples // args.batch_size, 1)
    num_images = 0
    for batch_i in range(num_batches):
        data = next(input_generator)  # this is batch_size'd
        results = model.predict(data, batch_size=args.batch_size)['texture_rgb']
        for i in range(results.shape[0]):
            result = results[i]
            img = keras_vgg_buddy.deprocess_image(result, contrast_percent=args.contrast_percent)
            imsave(args.output_prefix + '_{}.png'.format(num_images), img)
            num_images += 1


def generate_img_batches(root_path, batch_size, resize_shape=None):
    '''Yield random images loaded from `root_path`'''
    img_data = load_images(root_path, resize_shape=resize_shape)
    while True:
        imgs = np.array(random.sample(img_data, batch_size))
        yield imgs


def load_images(root_path, resize_shape=None):
    '''Load up and resize a directory full of images.'''
    data = []
    filenames = sorted(os.listdir(root_path))
    for filename in filenames:
        # TODO: improve vgg-buddy so this isn't necessary
        if resize_shape:
            width = resize_shape[-1]
            square = width == resize_shape[0]
        else:
            width = None
            square = False
        img = keras_vgg_buddy.load_and_preprocess_image(
            os.path.join(root_path, filename),
            width=width, square=square)
        data.append(img)
    return data



import os
import time

import numpy as np
import keras_vgg_buddy

from .generators.base import output_samples
from .generators.style_xfer import output_size_from_glob, transform_glob
from .generators.callbacks import GenerateSamplesCallback
from .models.config import get_model_by_name
from .training import train


def main(args):
    # ensure output path exists
    output_dir = os.path.dirname(args.output_prefix)
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    if args.train:
        train_main(args)
    else:
        generate_main(args)


def train_main(args):
    # should this allow training on rectangles or just squares?
    args.max_height = args.max_width
    assert args.style_image_path, 'Style image path required for training.'
    style_img = keras_vgg_buddy.load_and_preprocess_image(args.style_image_path, width=args.max_width, square=True)

    print('creating model...')
    model, input_generator, eval_generator = get_model_by_name(args.model)
    model = model(args, style_img=style_img)
    print('loading weights...')
    weights_filename = args.weights_prefix + '.weights'
    if not args.ignore_weights and os.path.exists(weights_filename):
        model.nodes['texnet'].load_weights(weights_filename)

    input_generator = input_generator(args)
    eval_generator = eval_generator(args)
    started_training = time.time()
    train(model, args, input_generator,
        callbacks=[GenerateSamplesCallback(model, args, eval_generator)]
    )
    print('Done training after {:.2f} seconds'.format(time.time() - started_training))
    print('saving weights')
    save_kwargs = {}
    if args.auto_save_weights:
        save_kwargs['overwrite'] = True
    model.nodes['texnet'].save_weights(weights_filename, **save_kwargs)
    # output final samples
    output_samples(model, args, eval_generator)


def generate_main(args):
    # determine input image size
    args.max_height, args.max_width = output_size_from_glob(args.convert_glob, width=args.max_width)
    print('creating model...')
    model, input_generator, eval_generator = get_model_by_name(args.model)
    model = model(args)
    print('loading weights...')
    weights_filename = args.weights_prefix + '.weights'
    if not args.ignore_weights and os.path.exists(weights_filename):
        model.nodes['texnet'].load_weights(weights_filename)

    transform_glob(model, args)


if __name__ == '__main__':
    import argparser
    import os

    args = argparser.get_parser()

    output_dir = os.path.dirname(args.output_prefix)
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    main(args)



import argparse
import os


VGG_ENV_VAR = 'VGG_WEIGHT_PATH'


class CommaSplitAction(argparse.Action):
    '''Split n strip incoming string argument.'''
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, [v.strip() for v in values.strip().split(',') if v.strip()])


def get_args():
    parser = argparse.ArgumentParser(description='Real-Time Style Transfer with Keras')
    parser.add_argument('output_prefix', metavar='output_prefix', type=str,
                        help='Prefix for output')
    parser.add_argument('--style-img', dest='style_image_path', type=str,
                        help='Path to the style image.')
    parser.add_argument('--max-width', dest='max_width', type=int,
                        default=256, help='Max width')
    parser.add_argument('--max-height', dest='max_height', type=int,
                        default=None, help='Max height. Just leave it alone for now.')
    parser.add_argument('--batch-size', dest='batch_size', type=int,
                        default=4, help='Batch size')
    parser.add_argument('--num-samples', dest='num_samples', type=int,
                        default=16, help='How many samples to generate every training epoch.')
    parser.add_argument('--weights-prefix', dest='weights_prefix', type=str,
                        default='texnet', help='Prefix for texturenet weight filenames')
    parser.add_argument('--ignore-weights', dest='ignore_weights', action='store_true',
                        help='Do not load any existing weights.')
    parser.add_argument('--auto-save-weights', dest='auto_save_weights', action='store_true',
                        help='Save over existing weights without asking.')
    parser.add_argument('--train', dest='train', action='store_true',
                        help='Train network')
    parser.add_argument('--convert-glob', dest='convert_glob', type=str,
                        default='', help='Glob for files to convert')
    parser.add_argument('--train-data', dest='training_data_path', type=str,
                        default='./training-images/', help='Path to training images.')
    parser.add_argument('--eval-data', dest='eval_data_path', type=str,
                        default='./training-images/', help='Path to evaluation images.')
    parser.add_argument('--contrast', dest='contrast_percent', type=float,
                        default=0.0, help='Stretch contrast percentile.')
    # losses
    parser.add_argument('--content-w', dest='content_weight', type=float,
                        default=1.0, help='Content loss weight')
    parser.add_argument('--content-layers', dest='content_layers', action=CommaSplitAction,
                        default=['conv2_2'],
                        help='Comma-separated list of layer names to be used for the content loss')
    parser.add_argument('--style-w', dest='style_weight', type=float,
                        default=10.0, help='Style loss weight')
    parser.add_argument('--style-layers', dest='style_layers', action=CommaSplitAction,
                        default=['conv1_2', 'conv2_2', 'conv3_3', 'conv4_3'],
                        help='Comma-separated list of layer names to be used for the content loss')
    parser.add_argument('--mrf-w', dest='mrf_weight', type=float,
                        default=0.0, help='MRF loss weight')
    parser.add_argument('--mrf-layers', dest='mrf_layers', action=CommaSplitAction,
                        default=['conv4_2'],
                        help='Comma-separated list of layer names to be used for the MRF loss')
    parser.add_argument('--tv-w', dest='tv_weight', type=float,
                        default=0.00001, help='Texture total variation loss (smoothness).')
    # vgg
    parser.add_argument('--vgg-weights', dest='vgg_weights', type=str,
                        default=os.environ.get(VGG_ENV_VAR, 'vgg16_weights.h5'), help='Path to VGG16 weights.')
    parser.add_argument('--pool-mode', dest='pool_mode', type=str,
                        default='max', help='Pooling mode for VGG ("avg" or "max")')
    # optimizer
    parser.add_argument('--learn-rate', dest='learn_rate', type=float,
                        default=0.1, help='Initial learning rate')
    parser.add_argument('--min-learn-rate', dest='min_learn_rate', type=float,
                        default=0.001, help='Final learning rate')
    parser.add_argument('--iters', dest='num_iterations_per_epoch', type=int,
                        default=100, help='Number of iterations per epoch')
    parser.add_argument('--num-epochs', dest='num_epochs', type=int,
                        default=400, help='Number of training epochs')
    parser.add_argument('--model', dest='model', type=str,
                        default='transfer', help='Which model to use. Only supports "transfer"')
    return parser.parse_args()



from keras.callbacks import Callback

from .base import output_samples


class GenerateSamplesCallback(Callback):
    '''Generate some samples every epoch.'''
    def __init__(self, model, args, input_generator, verbose=0):
        super(GenerateSamplesCallback, self).__init__()
        self.model = model
        self.args = args
        self.input_generator = input_generator
        self.verbose = verbose

    def on_epoch_end(self, epoch, logs={}):
        output_samples(self.model, self.args, self.input_generator)
